/****************************************************************************
 * This file is part of the project AqHome.
 * AqHome (c) by 2025 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/

#ifndef AQH_M_FLASHREADY_H
#define AQH_M_FLASHREADY_H


#include <aqhome/api.h>
#include <aqhome/ipc2/message.h>

#include <gwenhywfar/debug.h>


AQHOME_API uint32_t AQH_FlashReadyMessage_GetUid(const AQH_MESSAGE *msg);
AQHOME_API uint32_t AQH_FlashReadyMessage_GetManufacturer(const AQH_MESSAGE *msg);
AQHOME_API uint16_t AQH_FlashReadyMessage_GetDeviceType(const AQH_MESSAGE *msg);
AQHOME_API uint8_t AQH_FlashReadyMessage_GetDeviceVersion(const AQH_MESSAGE *msg);
AQHOME_API uint8_t AQH_FlashReadyMessage_GetDeviceRevision(const AQH_MESSAGE *msg);
AQHOME_API uint8_t AQH_FlashReadyMessage_GetFirmwareVariant(const AQH_MESSAGE *msg);
AQHOME_API uint8_t AQH_FlashReadyMessage_GetFirmwareVersionMajor(const AQH_MESSAGE *msg);
AQHOME_API uint8_t AQH_FlashReadyMessage_GetFirmwareVersionMinor(const AQH_MESSAGE *msg);
AQHOME_API uint8_t AQH_FlashReadyMessage_GetFirmwareVersionPatchlevel(const AQH_MESSAGE *msg);
AQHOME_API uint16_t AQH_FlashReadyMessage_GetPagesize(const AQH_MESSAGE *msg);

AQHOME_API void AQH_FlashReadyMessage_DumpToBuffer(const AQH_MESSAGE *msg, GWEN_BUFFER *dbuf, const char *sText);



#endif
